"""Test comic format."""
import shutil

from platform import system

from picopt import PROGRAM_NAME, cli
from tests import CONTAINER_DIR, get_test_dir


__all__ = ()
TMP_ROOT = get_test_dir()
FN = "test_zip.zip"
SRC_CBZ = CONTAINER_DIR / FN

if system() == "Darwin":
    FNS = {
        FN: (2974, 1917),
    }
else:
    FNS = {
        FN: (2974, 1861),
    }


class TestContainersDir:
    def setup_method(self) -> None:
        self.teardown_method()
        TMP_ROOT.mkdir(exist_ok=True)
        shutil.copy(SRC_CBZ, TMP_ROOT)
        for name, sizes in FNS.items():
            path = TMP_ROOT / name
            assert path.stat().st_size == sizes[0]

    def teardown_method(self) -> None:
        if TMP_ROOT.exists():
            shutil.rmtree(TMP_ROOT)

    def test_containers_no_convert(self) -> None:
        args = (PROGRAM_NAME, "-x", "ZIP", str(TMP_ROOT / FN))
        res = cli.main(args)
        assert res
        for name, sizes in FNS.items():
            path = TMP_ROOT / name
            assert path.stat().st_size == sizes[1]



"""Test comic format."""
import platform
import shutil

from datetime import datetime

from ruamel.yaml import YAML

from picopt import PROGRAM_NAME, cli
from tests import IMAGES_DIR, get_test_dir


__all__ = ()
TMP_ROOT = get_test_dir()
FN = "test_jpg.jpg"
SRC_JPG = IMAGES_DIR / FN
TMP_FN = str(TMP_ROOT / FN)
TIMESTAMPS_PATH = TMP_ROOT / ".picopt_treestamps.yaml"
WAL_PATH = TMP_ROOT / ".picopt_treestamps.wal.yaml"

if platform.system() == "Darwin":
    FNS = {
        FN: (97373, 87913),
    }
else:
    FNS = {
        FN: (97373, 87922),
    }

DEFAULT_CONFIG = {
    "bigger": False,
    "convert_to": [],
    "formats": ["GIF", "PNG", "JPEG", "WEBP"],
    "keep_metadata": True,
    "ignore": [],
    "recurse": True,
    "symlinks": True,
}


class TestContainersDir:
    @staticmethod
    def _assert_sizes(index, root=TMP_ROOT):
        for name, sizes in FNS.items():
            path = root / name
            assert path.stat().st_size == sizes[index]

    def setup_method(self) -> None:
        self.teardown_method()
        TMP_ROOT.mkdir(exist_ok=True)
        shutil.copy(SRC_JPG, TMP_ROOT)
        self._assert_sizes(0)

    def teardown_method(self) -> None:
        if TMP_ROOT.exists():
            shutil.rmtree(TMP_ROOT)

    @staticmethod
    def _write_timestamp(path, ts=None, config=None):
        if ts is None:
            ts = datetime.now().timestamp()
        if config is None:
            config = DEFAULT_CONFIG
        yaml = {"config": config, str(path): ts}
        YAML().dump(yaml, TIMESTAMPS_PATH)

    def test_no_timestamp(self) -> None:
        args = (PROGRAM_NAME, "-rtvv", TMP_FN)
        res = cli.main(args)
        assert res
        self._assert_sizes(1)

    def test_timestamp(self):
        self._write_timestamp(FN)
        args = (PROGRAM_NAME, "-rtvv", TMP_FN)
        res = cli.main(args)
        assert res
        self._assert_sizes(0)

    def test_different_config(self):
        self._write_timestamp(FN)
        args = (PROGRAM_NAME, "-brtvv", TMP_FN)
        res = cli.main(args)
        assert res
        self._assert_sizes(1)

    def test_timestamp_dir(self):
        self._write_timestamp(TMP_ROOT)
        args = (PROGRAM_NAME, "-rtvv", TMP_FN)
        res = cli.main(args)
        assert res
        self._assert_sizes(0)

    def _setup_child_dir(self):
        tmp_child_dir = TMP_ROOT / "child"
        tmp_child_dir.mkdir(exist_ok=True)
        shutil.copy(SRC_JPG, tmp_child_dir)
        self._assert_sizes(0, tmp_child_dir)
        return tmp_child_dir

    def test_timestamp_children(self):
        tmp_child_dir = self._setup_child_dir()
        self._write_timestamp(tmp_child_dir)
        args = (PROGRAM_NAME, "-rtvv", str(TMP_ROOT))
        res = cli.main(args)
        assert res
        self._assert_sizes(1)
        self._assert_sizes(0, tmp_child_dir)

    def test_timestamp_parents(self):
        tmp_child_dir = self._setup_child_dir()

        self._write_timestamp(TMP_ROOT)
        args = (PROGRAM_NAME, "-rtvv", str(tmp_child_dir))
        res = cli.main(args)
        assert res
        self._assert_sizes(0, tmp_child_dir)

    def test_journal_cleanup(self) -> None:
        args = (PROGRAM_NAME, "-rtvv", TMP_FN)
        res = cli.main(args)
        assert res
        assert not WAL_PATH.exists()

    @staticmethod
    def _write_wal(path, ts=None, config=None):
        if ts is None:
            ts = datetime.now().timestamp()
        if config is None:
            config = DEFAULT_CONFIG
        yaml = {"config": config, "wal": [{str(path): ts}]}
        YAML().dump(yaml, WAL_PATH)

    def test_timestamp_read_journal(self):
        self._write_wal(TMP_FN)
        args = (PROGRAM_NAME, "-rtvv", TMP_FN)
        res = cli.main(args)
        assert res
        self._assert_sizes(0)



"""Tests init."""
import inspect

from pathlib import Path


TEST_FILES_DIR = Path("tests/test_files")
IMAGES_DIR = TEST_FILES_DIR / "images"
INVALID_DIR = TEST_FILES_DIR / "invalid"
CONTAINER_DIR = TEST_FILES_DIR / "containers"
TMP_ROOT = "/tmp"


def get_test_dir():
    """Return a module specific tmpdir."""
    frame = inspect.currentframe()
    if frame and frame.f_back:
        caller = frame.f_back
        module_name = caller.f_globals["__name__"]
    else:
        module_name = "unknown"

    return TMP_ROOT / Path("picopt-" + module_name)



"""Handler for zip files."""
import os

from pathlib import Path
from typing import Optional, Union
from zipfile import ZIP_DEFLATED, ZipFile, is_zipfile

from PIL import Image, UnidentifiedImageError
from rarfile import RarFile, is_rarfile

from picopt.handlers.container import ContainerHandler
from picopt.handlers.handler import Format


class Zip(ContainerHandler):
    """Ziplike container."""

    OUTPUT_FORMAT: str = "ZIP"
    OUTPUT_FORMAT_OBJ: Format = Format(OUTPUT_FORMAT)
    INPUT_FORMAT_RAR: str = "RAR"
    INPUT_FORMAT_OBJ_RAR: Format = Format(INPUT_FORMAT_RAR)
    RAR_SUFFIX: str = "." + INPUT_FORMAT_RAR.lower()
    PROGRAMS: dict[str, Optional[str]] = {ContainerHandler.INTERNAL: None}

    @classmethod
    def identify_format(cls, path: Path) -> Optional[Format]:
        """Return the format if this handler can handle this path."""
        format = None
        suffix = path.suffix.lower()
        if is_zipfile(path) and suffix == cls.output_suffix():
            format = cls.OUTPUT_FORMAT_OBJ
        elif is_rarfile(path) and suffix == cls.RAR_SUFFIX:
            format = cls.INPUT_FORMAT_OBJ_RAR
        return format

    def _get_archive(self) -> Union[ZipFile, RarFile]:
        """Use the zipfile builtin for this archive."""
        if is_zipfile(self.original_path):
            archive = ZipFile(self.original_path, "r")
        elif is_rarfile(self.original_path):
            archive = RarFile(self.original_path, "r")
        else:
            raise ValueError(f"Unknown archive type: {self.original_path}")
        return archive

    def _set_comment(self, comment: Union[str, bytes, None]) -> None:
        """Set the comment from the archive."""
        if type(comment) is str:
            self.comment = comment.encode()
        elif type(comment) is bytes:
            self.comment = comment

    def unpack_into(self) -> None:
        """Uncompress archive."""
        with self._get_archive() as archive:
            archive.extractall(self.tmp_container_dir)
            self._set_comment(archive.comment)

    @staticmethod
    def _is_image(path: Path) -> bool:
        """Is a file an image."""
        result = False
        try:
            with Image.open(path, mode="r") as image:
                if image.format:
                    result = True
            image.close()
        except UnidentifiedImageError:
            pass
        return result

    def pack_into(self, working_path: Path) -> None:
        """Zip up the files in the tempdir into the new filename."""
        with ZipFile(
            working_path, "w", compression=ZIP_DEFLATED, compresslevel=9
        ) as new_zf:
            for root, _, filenames in os.walk(self.tmp_container_dir):
                root_path = Path(root)
                for fname in sorted(filenames):
                    if self.config.verbose:
                        print(".", end="")
                    full_path = root_path / fname
                    if self._is_image(full_path):
                        # Do not deflate images in zipfile.
                        # Picopte should have already achieved maximum
                        # compression over deflate.
                        compress_type = None
                    else:
                        compress_type = ZIP_DEFLATED
                    archive_path = full_path.relative_to(self.tmp_container_dir)
                    new_zf.write(full_path, archive_path, compress_type)
            if self.comment:
                new_zf.comment = self.comment


class CBZ(Zip):
    """CBZ Container."""

    OUTPUT_FORMAT: str = "CBZ"
    OUTPUT_FORMAT_OBJ: Format = Format(OUTPUT_FORMAT)
    INPUT_FORMAT_RAR: str = "CBR"
    INPUT_FORMAT_OBJ_RAR: Format = Format(INPUT_FORMAT_RAR)
    RAR_SUFFIX: str = "." + INPUT_FORMAT_RAR.lower()


class EPub(Zip):
    """Epub Container."""

    OUTPUT_FORMAT: str = "EPUB"
    OUTPUT_FORMAT_OBJ: Format = Format(OUTPUT_FORMAT)
    CONVERT: bool = False



"""Format Superclass."""
from abc import ABCMeta
from pathlib import Path
from typing import Any

from PIL import Image
from PIL.BmpImagePlugin import BmpImageFile
from PIL.PpmImagePlugin import PpmImageFile
from PIL.TiffImagePlugin import TiffImageFile

from picopt.handlers.handler import Format, Handler
from picopt.stats import ReportStats


PPM_FORMAT_OBJ = Format(PpmImageFile.format, True, False)
BPM_FORMAT_OBJ = Format(BmpImageFile.format, True, False)
CONVERTABLE_FORMAT_OBJS = set((BPM_FORMAT_OBJ, PPM_FORMAT_OBJ))
CONVERTABLE_FORMATS = set([format.format for format in CONVERTABLE_FORMAT_OBJS])
TIFF_FORMAT = TiffImageFile.format
TIFF_FORMAT_OBJ = Format(TIFF_FORMAT, True, False)
TIFF_ANIMATED_FORMAT_OBJ = Format(TIFF_FORMAT, True, True)


class ImageHandler(Handler, metaclass=ABCMeta):
    """Format superclass."""

    PIL2_ARGS: dict[str, Any] = {}
    PREFERRED_PROGRAM: str = "unimplemented"

    def _optimize_with_progs(self) -> ReportStats:
        """
        Use the correct optimizing functions in sequence.

        And report back statistics.
        """
        path = self.original_path
        for func in self.PROGRAMS.keys():
            if path != self.original_path:
                self.working_paths.add(path)
            new_path = self.get_working_path(func)
            path = getattr(self, func)(path, new_path)
            if self.BEST_ONLY:
                break

        bytes_count = self.cleanup_after_optimize(path)
        report_stats = ReportStats(self.final_path, bytes_count=bytes_count)

        return report_stats

    def optimize_image(self) -> ReportStats:
        """Optimize a given image from a filename."""
        try:
            report_stats = self._optimize_with_progs()
            if self.config.verbose:
                report_stats.report(self.config.test)
        except Exception as exc:
            report_stats = self.error(exc)
        return report_stats

    def pil2native(self, old_path: Path, new_path: Path) -> Path:
        """Use PIL to save the image."""
        if (
            self.input_format_obj == TIFF_FORMAT_OBJ
            or self.PREFERRED_PROGRAM not in self.config._available_programs
        ):
            with Image.open(old_path) as image:
                image.save(
                    new_path,
                    self.OUTPUT_FORMAT,
                    exif=self.metadata.exif,
                    icc_profile=self.metadata.icc_profile,
                    **self.PIL2_ARGS,
                )
            image.close()  # for animated images
        else:
            new_path = old_path
        return new_path



"""Test cli module."""
import shutil
import sys

from picopt import cli
from picopt.handlers.zip import CBZ, Zip
from tests import IMAGES_DIR, get_test_dir


__all__ = ()  # hides module from pydocstring
TYPE_NAME = "png"
TMP_ROOT = get_test_dir()
JPEG_SRC = IMAGES_DIR / "test_jpg.jpg"


class TestCLI:
    OLD_PATH = TMP_ROOT / "old.jpg"

    def setup_method(self) -> None:
        TMP_ROOT.mkdir(exist_ok=True)

    def teardown_method(self) -> None:
        if TMP_ROOT.exists():
            shutil.rmtree(TMP_ROOT)

    def test_get_arguments(self) -> None:
        args = ("picopt", "-rqc", "PNG,WEBP", "-x", "CBZ,ZIP", "-bTLM", str(TMP_ROOT))
        arguments = cli.get_arguments(args)
        arguments = arguments.picopt
        assert arguments.verbose == 0
        assert ("PNG", "WEBP") == arguments.convert_to
        assert arguments.formats is None
        assert arguments._extra_formats == (CBZ.OUTPUT_FORMAT, Zip.OUTPUT_FORMAT)
        assert arguments.symlinks
        assert arguments.bigger
        assert not arguments.timestamps
        assert arguments.test
        assert arguments.list_only
        assert not arguments.keep_metadata
        assert arguments.paths[0] == str(TMP_ROOT)

    def test_main(self) -> None:
        shutil.copy(JPEG_SRC, self.OLD_PATH)
        old_size = self.OLD_PATH.stat().st_size
        sys.argv = ["picopt", str(self.OLD_PATH)]
        cli.main()
        assert self.OLD_PATH.is_file()
        assert self.OLD_PATH.stat().st_size < old_size

    def test_main_err(self) -> None:
        sys.argv = ["picopt", "XXX"]
        try:
            cli.main()
        except SystemExit as exc:
            assert exc.code == 1



"""Test png bit depth utility module."""
import sys

from picopt.pillow.png_bit_depth import main, png_bit_depth
from tests import IMAGES_DIR


__all__ = ()  # hides module from pydocstring
TEST_SRC_PATH = IMAGES_DIR / "test_png.png"
TEST_SRC_PATH_16 = IMAGES_DIR / "test_png_16rgba.png"
TEST_SRC_PATH_JPG = IMAGES_DIR / "test_jpg.jpg"


def test_png_bit_depth() -> None:
    res = png_bit_depth(TEST_SRC_PATH)
    assert res == 8


def test_png_bit_depth_16() -> None:
    res = png_bit_depth(TEST_SRC_PATH_16)
    assert res == 16


def test_png_bit_depth_invalid() -> None:
    res = png_bit_depth(TEST_SRC_PATH_JPG)
    assert res is None


def test_main() -> None:
    sys.argv[1] = str(TEST_SRC_PATH)
    main()



#!/usr/bin/env python3
"""
Determine if a webp is lossless.

This should be a part of Pillow
https://developers.google.com/speed/webp/docs/webp_lossless_bitstream_specification
"""
from pathlib import Path

from picopt.pillow.unpack import compare_header


RIFF_HEADER: tuple[bytes, ...] = (b"R", b"I", b"F", b"F")
WEBP_HEADER: tuple[bytes, ...] = (b"W", b"E", b"B", b"P")
VP8L_HEADER: tuple[bytes, ...] = (b"V", b"P", b"8", b"L")

COMPARATORS = ((0, RIFF_HEADER), (8, WEBP_HEADER), (12, VP8L_HEADER))


def is_lossless(filename: str) -> bool:
    """If a file is a png, get the bit depth from the standard position."""
    path = Path(filename)
    try:
        with path.open("rb") as img:
            for seek, byte_array in COMPARATORS:
                compare_header(img, seek, byte_array)
        return True
    except ValueError:
        return False


def main() -> None:
    """Stand alone cli tool for getting lossless status."""
    import sys

    lossless = is_lossless(sys.argv[1])
    print(lossless)


if __name__ == "__main__":
    main()



"""Statistics for the optimization operations."""
from pathlib import Path
from typing import Optional

from humanize import naturalsize
from termcolor import cprint


class ReportStats:
    """Container for reported stats from optimization operations."""

    def __init__(
        self,
        path: Path,
        bytes_count: Optional[tuple[int, int]] = None,
        error: Optional[str] = None,
    ) -> None:
        """Initialize required instance variables."""
        self.path = path
        self.error = error
        if bytes_count:
            self.bytes_in = bytes_count[0]
            self.bytes_out = bytes_count[1]
        else:
            self.bytes_in = 0
            self.bytes_out = 0

    def _new_percent_saved(self) -> str:
        """Spit out how much space the optimization saved."""
        size_in = self.bytes_in
        size_out = self.bytes_out
        if size_in <= 0:
            ratio = 1.0
        else:
            ratio = size_out / size_in
        saved = naturalsize(size_in - size_out)
        percent_saved = (1 - ratio) * 100

        result = "{:.{prec}f}% ({})".format(percent_saved, saved, prec=2)
        return result

    def _report_saved(self, test: bool) -> str:
        """Return the percent saved."""
        report = ""

        report += f"{self.path}: "
        report += self._new_percent_saved()
        if test:
            report += " could be saved."
        return report

    def report(self, test: bool, color="white") -> None:
        """Record the percent saved & print it."""
        if self.error:
            report = f"{self.path} error: {self.error}"
        else:
            report = self._report_saved(test)
        cprint(report, color)



"""Confuse config for picopt."""
import pathlib
import stat
import time
import typing

from argparse import Namespace
from copy import deepcopy

from confuse import Configuration
from confuse.templates import (
    AttrDict,
    Choice,
    Integer,
    MappingTemplate,
    Optional,
    Path,
    Sequence,
)
from dateutil.parser import parse
from termcolor import cprint

from picopt import PROGRAM_NAME
from picopt.handlers.gif import AnimatedGif, Gif
from picopt.handlers.handler import Format, Handler
from picopt.handlers.image import (
    BPM_FORMAT_OBJ,
    CONVERTABLE_FORMAT_OBJS,
    CONVERTABLE_FORMATS,
    PPM_FORMAT_OBJ,
    TIFF_ANIMATED_FORMAT_OBJ,
    TIFF_FORMAT,
    TIFF_FORMAT_OBJ,
)
from picopt.handlers.jpeg import Jpeg
from picopt.handlers.png import Png
from picopt.handlers.webp import Gif2WebP, WebPLossless, WebPLossy
from picopt.handlers.webp_animated import WebPAnimated
from picopt.handlers.zip import CBZ, EPub, Zip


_PNG_CONVERTABLE_FORMAT_OBJS = CONVERTABLE_FORMAT_OBJS | frozenset(
    [Gif.OUTPUT_FORMAT_OBJ]
)
_WEBP_CONVERTABLE_FORMAT_OBJS = _PNG_CONVERTABLE_FORMAT_OBJS | frozenset(
    [Png.OUTPUT_FORMAT_OBJ]
)
PNG_CONVERTABLE_FORMATS = frozenset(
    [format.format for format in _PNG_CONVERTABLE_FORMAT_OBJS]
)
WEBP_CONVERTABLE_FORMATS = frozenset(
    [format.format for format in _WEBP_CONVERTABLE_FORMAT_OBJS]
)
CONVERT_TO_FORMATS = frozenset(
    (
        Png.OUTPUT_FORMAT,
        WebPLossless.OUTPUT_FORMAT,
        Zip.OUTPUT_FORMAT,
        CBZ.OUTPUT_FORMAT,
    )
)
CONTAINER_CONVERTABLE_FORMATS = frozenset((Zip.INPUT_FORMAT_RAR, CBZ.INPUT_FORMAT_RAR))
DEFAULT_HANDLERS = (Gif, AnimatedGif, Jpeg, Png, WebPLossy, WebPLossless)
HANDLERS = frozenset([*DEFAULT_HANDLERS, Gif2WebP, WebPAnimated, Zip, CBZ, EPub])
ALL_FORMATS: frozenset[str] = (
    frozenset([cls.OUTPUT_FORMAT for cls in HANDLERS])
    | CONVERTABLE_FORMATS
    | frozenset([TIFF_FORMAT])
    | CONTAINER_CONVERTABLE_FORMATS
)
TEMPLATE = MappingTemplate(
    {
        "picopt": MappingTemplate(
            {
                "after": Optional(float),
                "bigger": bool,
                "convert_to": Optional(Sequence(Choice(CONVERT_TO_FORMATS))),
                "keep_metadata": bool,
                "formats": Sequence(Choice(ALL_FORMATS)),
                "ignore": Sequence(str),
                "jobs": Integer(),
                "list_only": bool,
                "paths": Sequence(Path()),
                "recurse": bool,
                "symlinks": bool,
                "test": bool,
                "timestamps": bool,
                "verbose": Integer(),
                "_available_programs": set,
                "_extra_formats": Optional(Sequence(Choice(ALL_FORMATS))),
                "_format_handlers": dict,
            }
        )
    }
)
TIMESTAMPS_CONFIG_KEYS = set(
    (
        "bigger",
        "convert_to",
        "formats",
        "ignore",
        "keep_metadata",
        "recurse",
        "symlinks",
    )
)
# Handlers for formats are listed in priority order
_FORMAT_HANDLERS = {
    PPM_FORMAT_OBJ: {"convert": (WebPLossless, Png)},
    BPM_FORMAT_OBJ: {"convert": (WebPLossless, Png)},
    Gif.OUTPUT_FORMAT_OBJ: {"convert": (Gif2WebP, Png), "native": (Gif,)},
    AnimatedGif.OUTPUT_FORMAT_OBJ: {"convert": (Gif2WebP,), "native": (AnimatedGif,)},
    Jpeg.OUTPUT_FORMAT_OBJ: {"native": (Jpeg,)},
    Png.OUTPUT_FORMAT_OBJ: {"convert": (WebPLossless,), "native": (Png,)},
    WebPLossy.OUTPUT_FORMAT_OBJ: {"native": (WebPLossy,)},
    WebPLossless.OUTPUT_FORMAT_OBJ: {"native": (WebPLossless,)},
    WebPAnimated.OUTPUT_FORMAT_OBJ: {"native": (WebPAnimated,)},
    Zip.OUTPUT_FORMAT_OBJ: {"native": (Zip,)},
    CBZ.OUTPUT_FORMAT_OBJ: {"native": (CBZ,)},
    Zip.INPUT_FORMAT_OBJ_RAR: {"convert": (Zip,)},
    CBZ.INPUT_FORMAT_OBJ_RAR: {"convert": (CBZ,)},
    EPub.OUTPUT_FORMAT_OBJ: {"native": (EPub,)},
    TIFF_FORMAT_OBJ: {"convert": (WebPLossless, Png)},
    TIFF_ANIMATED_FORMAT_OBJ: {"convert": (WebPAnimated,)},
}
MODE_EXECUTABLE = stat.S_IXUSR ^ stat.S_IXGRP ^ stat.S_IXOTH


def _is_external_program_executable(
    program: str, bin_path: typing.Optional[str], verbose: int
) -> bool:
    """Test to see if the external programs can be run."""
    try:
        if not bin_path:
            raise ValueError()
        path = pathlib.Path(bin_path)
        mode = path.stat().st_mode
        result = bool(mode & MODE_EXECUTABLE)
    except Exception:
        if bin_path and verbose:
            cprint(
                f"WARNING: Could not find executable program for {program}", "yelllow"
            )
        result = False

    return result


def _get_available_programs(config: Configuration) -> set:
    """Run the external program tester on the required binaries."""
    verbose = config["picopt"]["verbose"].get(int)
    if not isinstance(verbose, int):
        raise ValueError(f"wrong type for convert_to: {type(verbose)} {verbose}")
    programs = set()
    for handler in HANDLERS:
        for program, bin_path in handler.PROGRAMS.items():
            if (
                program == Handler.INTERNAL
                or program.startswith("pil2")
                or _is_external_program_executable(program, bin_path, verbose)
            ):
                programs.add(program)
    if not programs:
        raise ValueError("No optimizers are available or all optimizers are disabled")
    return programs


def _config_list_to_set(config, key) -> set[str]:
    val_list = config["picopt"][key].get(list)
    if not isinstance(val_list, (tuple, list, set)):
        raise ValueError(f"wrong type for convert_to: {type(val_list)} {val_list}")
    return set(val_list)


def _update_formats(config: Configuration) -> dict:
    formats = _config_list_to_set(config, "formats")
    if "_extra_formats" in config["picopt"]:
        formats |= _config_list_to_set(config, "_extra_formats")

    convert_to = _config_list_to_set(config, "convert_to")
    convert_handlers: dict[typing.Type[Handler], set[Format]] = {}
    if Png.OUTPUT_FORMAT in convert_to:
        formats |= PNG_CONVERTABLE_FORMATS
        format_objs = deepcopy(_PNG_CONVERTABLE_FORMAT_OBJS)
        if TIFF_FORMAT in formats:
            format_objs.add(TIFF_FORMAT_OBJ)
        convert_handlers[Png] = format_objs
    if WebPLossless.OUTPUT_FORMAT in convert_to:
        formats |= WEBP_CONVERTABLE_FORMATS
        format_objs = deepcopy(_WEBP_CONVERTABLE_FORMAT_OBJS)
        if TIFF_FORMAT in formats:
            format_objs.add(TIFF_FORMAT_OBJ)
        convert_handlers[WebPLossless] = format_objs

        convert_handlers[Gif2WebP] = set(
            [Gif.OUTPUT_FORMAT_OBJ, AnimatedGif.OUTPUT_FORMAT_OBJ]
        )
        if TIFF_FORMAT in formats:
            convert_handlers[WebPAnimated] = set([TIFF_ANIMATED_FORMAT_OBJ])
    if Zip.OUTPUT_FORMAT in convert_to:
        formats |= set([Zip.INPUT_FORMAT_RAR])
        convert_handlers[Zip] = set([Zip.INPUT_FORMAT_OBJ_RAR])
    if CBZ.OUTPUT_FORMAT in convert_to:
        formats |= set([CBZ.INPUT_FORMAT_RAR])
        convert_handlers[CBZ] = set([CBZ.INPUT_FORMAT_OBJ_RAR])

    config["picopt"]["formats"].set(sorted(formats))

    return convert_handlers


def _create_format_handler_map(config: Configuration, convert_handlers: dict) -> None:
    """Create a format to handler map from config."""
    available_programs = _get_available_programs(config)
    format_handlers = {}
    formats = config["picopt"]["formats"].get(list)
    if not isinstance(formats, list):
        raise ValueError(f"wrong type for formats: {type(formats)}{formats}")
    formats = set(formats)

    for format, possible_handler_classes_group in _FORMAT_HANDLERS.items():
        if format.format not in formats:
            continue
        for (
            handler_type,
            possible_handler_classes,
        ) in possible_handler_classes_group.items():
            for handler_class in possible_handler_classes:
                available = handler_class.is_handler_available(
                    convert_handlers, available_programs, format
                )
                if available:
                    if format not in format_handlers:
                        format_handlers[format] = {}
                    format_handlers[format][handler_type] = handler_class
                    break
    config["picopt"]["_format_handlers"].set(format_handlers)
    config["picopt"]["_available_programs"].set(available_programs)


def _set_after(config) -> None:
    after = config["picopt"]["after"].get()
    if after is None:
        return

    try:
        timestamp = float(after)
    except ValueError:
        after_dt = parse(after)
        timestamp = time.mktime(after_dt.timetuple())

    config["picopt"]["after"].set(timestamp)


def _set_ignore(config) -> None:
    """Remove duplicates from the ignore list."""
    ignore_list: list[str] = config["picopt"]["ignore"].get(list)
    config["picopt"]["ignore"].set(tuple(sorted(set(ignore_list))))


def _set_timestamps(config) -> None:
    """Set the timestamps attribute."""
    timestamps = (
        config["picopt"]["timestamps"].get(bool)
        and not config["picopt"]["test"].get(bool)
        and not config["picopt"]["list_only"].get(bool)
    )
    config["picopt"]["timestamps"].set(timestamps)


def get_config(
    args: typing.Optional[Namespace] = None, modname=PROGRAM_NAME
) -> AttrDict:
    """Get the config dict, layering env and args over defaults."""
    config = Configuration(PROGRAM_NAME, modname=modname, read=False)
    try:
        config.read()
    except Exception as exc:
        cprint(f"WARNING: {exc}")
    if args and args.picopt and args.picopt.config:
        config.set_file(args.picopt.config)
    config.set_env()
    if args:
        config.set_args(args)
    _set_after(config)
    convert_handlers = _update_formats(config)
    _create_format_handler_map(config, convert_handlers)
    _set_ignore(config)
    _set_timestamps(config)
    ad = config.get(TEMPLATE)
    if not isinstance(ad, AttrDict):
        raise ValueError()
    return ad.picopt



"""JPEG format."""
from copy import copy
from pathlib import Path
from typing import Optional

from PIL.JpegImagePlugin import JpegImageFile

from picopt.handlers.handler import Format
from picopt.handlers.image import ImageHandler


class Jpeg(ImageHandler):
    """JPEG format class."""

    OUTPUT_FORMAT = JpegImageFile.format
    OUTPUT_FORMAT_OBJ = Format(OUTPUT_FORMAT, False, False)
    PROGRAMS: dict[str, Optional[str]] = ImageHandler.init_programs(
        ("mozjpeg", "jpegtran")
    )
    _ARGS_PREFIX = ["-optimize", "-progressive", "-copy"]
    _MOZJPEG_ARGS_PREFIX = [PROGRAMS["mozjpeg"], *_ARGS_PREFIX]
    _JPEGTRAN_ARGS_PREFIX = [PROGRAMS["jpegtran"], *_ARGS_PREFIX]

    @classmethod
    def _output_suffix(cls) -> str:
        """JPEG suffix does not match format."""
        return "jpg"

    def _jpegtran(
        self, args: list[Optional[str]], old_path: Path, new_path: Path
    ) -> Path:
        """Run the jpegtran type program."""
        args = copy(args)
        if not bin:
            return old_path
        if self.config.keep_metadata:
            args += ["all"]
        else:
            args += ["none"]
        args += ["-outfile", str(new_path), str(old_path)]
        args_t = tuple(args)
        self.run_ext(args_t)  # type: ignore
        return new_path

    def mozjpeg(self, old_path: Path, new_path: Path) -> Path:
        """Create argument list for mozjpeg."""
        return self._jpegtran(self._MOZJPEG_ARGS_PREFIX, old_path, new_path)

    def jpegtran(self, old_path: Path, new_path: Path) -> Path:
        """Create argument list for jpegtran."""
        return self._jpegtran(self._JPEGTRAN_ARGS_PREFIX, old_path, new_path)



#!/usr/bin/env python3
"""Run pictures through image specific external optimizers."""
import argparse

from argparse import Action, Namespace
from importlib.metadata import PackageNotFoundError, version
from typing import Optional

from picopt import PROGRAM_NAME, walk
from picopt.config import ALL_FORMATS, DEFAULT_HANDLERS, get_config
from picopt.handlers.png import Png
from picopt.handlers.webp import WebP
from picopt.handlers.zip import CBZ, Zip


DEFAULT_FORMATS = frozenset(
    [handler_cls.OUTPUT_FORMAT for handler_cls in DEFAULT_HANDLERS]
)
EXTRA_FORMATS = ALL_FORMATS - DEFAULT_FORMATS
FORMAT_DELIMETER = ","
try:
    VERSION = version(PROGRAM_NAME)
except PackageNotFoundError:
    VERSION = "test"


class SplitArgsAction(Action):
    """Convert csv string from argparse to a list."""

    def __call__(self, parser, namespace, values, _option_string=None):
        """Split values string into list."""
        if isinstance(values, str):
            values = tuple(sorted(values.strip().split(FORMAT_DELIMETER)))
        setattr(namespace, self.dest, values)


def _comma_join(formats: frozenset[str]) -> str:
    """Sort and join a sequence into a human readable string."""
    return ", ".join(sorted(formats))


def get_arguments(params: Optional[tuple[str, ...]] = None) -> Namespace:
    """Parse the command line."""
    description = "Losslessly optimizes and optionally converts images."
    parser = argparse.ArgumentParser(description=description)
    ###########
    # Options #
    ###########
    parser.add_argument(
        "-r",
        "--recurse",
        action="store_true",
        dest="recurse",
        help="Recurse down through directories on the command line.",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="count",
        dest="verbose",
        help="Display more output. Can be used multiple times for "
        "increasingly noisy output.",
    )
    parser.add_argument(
        "-q",
        "--quiet",
        action="store_const",
        dest="verbose",
        const=0,
        help="Display little to no output",
    )
    parser.add_argument(
        "-f",
        "--formats",
        action=SplitArgsAction,
        dest="formats",
        help="Only optimize images of the specified "
        f"'{FORMAT_DELIMETER}' delimited formats from: {_comma_join(ALL_FORMATS)}. "
        f"Defaults to {_comma_join(DEFAULT_FORMATS)}",
    )
    parser.add_argument(
        "-x",
        "--extra-formats",
        action=SplitArgsAction,
        dest="_extra_formats",
        help="Append additional formats to the default formats.",
    )
    parser.add_argument(
        "-c",
        "--convert-to",
        action=SplitArgsAction,
        dest="convert_to",
        help="A list of formats to convert to. Lossless images may convert to "
        f"{Png.OUTPUT_FORMAT} or {WebP.OUTPUT_FORMAT}. {Zip.INPUT_FORMAT_RAR} archives "
        f"may convert to {Zip.OUTPUT_FORMAT} or {CBZ.OUTPUT_FORMAT}. "
        "By default formats are not converted to other formats.",
    )
    parser.add_argument(
        "-S",
        "--no-symlinks",
        action="store_false",
        dest="symlinks",
        help="Do not follow symlinks for files and directories",
    )
    parser.add_argument(
        "-i",
        "--ignore",
        action=SplitArgsAction,
        dest="ignore",
        help="List of globs to ignore.",
    )
    parser.add_argument(
        "-b",
        "--bigger",
        action="store_true",
        dest="bigger",
        help="Save optimized files that are larger than the originals",
    )
    parser.add_argument(
        "-t",
        "--timestamps",
        action="store_true",
        dest="timestamps",
        help="Record the optimization time in a timestamps file. "
        "Do not optimize files that are older than their timestamp record.",
    )
    parser.add_argument(
        "-A",
        "--after",
        action="store",
        dest="after",
        help="Only optimize files after the specified timestamp. "
        "Supersedes recorded timestamp files. Can be an epoch number or "
        "datetime string",
    )
    parser.add_argument(
        "-T",
        "--test",
        action="store_true",
        dest="test",
        help="Report how much would be saved, but do not replace files.",
    )
    parser.add_argument(
        "-L",
        "--list",
        action="store_true",
        dest="list_only",
        help="Only list files that would be optimized",
    )

    parser.add_argument(
        "-M",
        "--destroy-metadata",
        action="store_false",
        dest="keep_metadata",
        help="Destroy metadata like EXIF and ICC Profiles",
    )
    parser.add_argument(
        "-j",
        "--jobs",
        type=int,
        action="store",
        dest="jobs",
        help="Number of parallel jobs to run simultaneously. Defaults "
        "to maximum available.",
    )
    parser.add_argument(
        "-C",
        "--config",
        type=str,
        action="store",
        help="Path to a config file",
    )
    ###########
    # Actions #
    ###########
    parser.add_argument(
        "-V",
        "--version",
        action="version",
        version=VERSION,
        help="Display the version number",
    )
    ###########
    # Targets #
    ###########
    parser.add_argument(
        "paths",
        metavar="path",
        type=str,
        nargs="+",
        help="File or directory paths to optimize",
    )

    if params is not None:
        params = params[1:]

    pns = parser.parse_args(params)
    return Namespace(picopt=pns)


def main(args: Optional[tuple[str, ...]] = None) -> bool:
    """Process command line arguments and walk inputs."""
    arguments = get_arguments(args)

    config = get_config(arguments)
    wob = walk.Walk(config)
    return wob.run()


if __name__ == "__main__":
    main()



"""Walk the directory trees and files and call the optimizers."""
import shutil
import time
import traceback

from multiprocessing.pool import ApplyResult, Pool
from pathlib import Path
from queue import SimpleQueue
from typing import Any, Optional, Type, Union

from confuse.templates import AttrDict
from humanize import naturalsize
from termcolor import cprint
from treestamps import Treestamps

from picopt import PROGRAM_NAME
from picopt.config import (
    PNG_CONVERTABLE_FORMATS,
    TIMESTAMPS_CONFIG_KEYS,
    WEBP_CONVERTABLE_FORMATS,
)
from picopt.handlers.container import ContainerHandler
from picopt.handlers.factory import create_handler
from picopt.handlers.handler import Handler
from picopt.handlers.image import TIFF_FORMAT, ImageHandler
from picopt.handlers.png import Png
from picopt.handlers.webp import WebP
from picopt.handlers.zip import CBZ, Zip
from picopt.old_timestamps import OldTimestamps
from picopt.stats import ReportStats
from picopt.tasks import (
    CompleteContainerTask,
    CompleteDirTask,
    CompleteTask,
    ContainerResult,
    DirResult,
    Totals,
)


class Walk:
    """Walk object for storing state of a walk run."""

    TIMESTAMPS_FILENAMES = set(Treestamps.get_filenames(PROGRAM_NAME))

    def __init__(self, config: AttrDict) -> None:
        """Initialize."""
        self._config: AttrDict = config
        top_paths = []
        for path in sorted(set(self._config.paths)):
            if path.is_symlink() and not self._config.symlinks:
                continue
            top_paths.append(path)
        self._top_paths: tuple[Path, ...] = tuple(top_paths)
        self._timestamps: dict[Path, Treestamps] = {}
        self._queues: dict[Path, SimpleQueue[Any]] = {}
        if self._config.jobs:
            self._pool = Pool(self._config.jobs)
        else:
            self._pool = Pool()

    def _is_skippable(self, path: Path) -> bool:
        """Handle things that are not optimizable files."""
        # File types
        if not self._config.symlinks and path.is_symlink():
            if self._config.verbose > 1:
                cprint(f"Skip symlink {path}", "white", attrs=["dark"])
            return True
        elif path.name in self.TIMESTAMPS_FILENAMES:
            return True
        elif not path.exists():
            if self._config.verbose > 1:
                cprint(f"WARNING: {path} not found.", "yellow")
            return True

        skip = False
        for ignore_glob in self._config.ignore:
            if path.match(ignore_glob):
                skip = True
                break

        return skip

    def _is_older_than_timestamp(
        self,
        path: Path,
        top_path: Path,
        container_mtime: Optional[float],
    ) -> bool:
        """Is the file older than the timestamp."""
        # If the file is in an container, use the container time.
        # This helps if you have a new container that you
        # collected from someone who put really old files in it that
        # should still be optimised
        if container_mtime is not None:
            mtime = container_mtime
        else:
            mtime = path.stat().st_mtime

        # The timestamp or configured walk after time for comparison.
        walk_after = None
        if self._config.after is not None:
            walk_after = self._config.after
        elif container_mtime is None:
            walk_after = self._timestamps.get(top_path, {}).get(path)

        if walk_after is None:
            return False

        return bool(mtime <= walk_after)

    def _clean_up_working_files(self, path):
        """Auto-clean old working temp files if encountered."""
        try:
            shutil.rmtree(path, ignore_errors=True)
            if self._config.verbose > 1:
                print(f"Deleted {path}")
        except Exception as exc:
            cprint(str(exc), "red")

    def walk_file(
        self,
        path: Path,
        top_path: Path,
        container_mtime: Optional[float] = None,
        convert: bool = True,
    ) -> Union[ApplyResult, DirResult, None]:
        """Optimize an individual file."""
        result: Union[ApplyResult, DirResult, None] = None
        try:
            if self._is_skippable(path):
                return result

            if path.name.rfind(Handler.WORKING_SUFFIX) > -1:
                self._clean_up_working_files(path)
                return result

            if path.is_dir():
                if self._config.recurse or container_mtime is not None:
                    result = self.walk_dir(
                        path, top_path, container_mtime, convert=convert
                    )
                return result

            if self._is_older_than_timestamp(path, top_path, container_mtime):
                return result

            handler = create_handler(self._config, path, convert=convert)

            if handler is None:
                return result

            if self._config.list_only:
                print(f"{path}: {handler.__class__.__name__}")
                return result

            if isinstance(handler, ContainerHandler):
                # Unpack inline, not in the pool, and walk immediately like dirs.
                handler.unpack()
                result = self._walk_container_dir(top_path, handler)
            elif isinstance(handler, ImageHandler):
                result = self._pool.apply_async(handler.optimize_image)
            else:
                raise ValueError(f"bad handler {handler}")
        except Exception as exc:
            traceback.print_exc()

            result = self._pool.apply_async(ReportStats, args=(path, None, exc))
        return result

    def walk_dir(
        self,
        dir_path: Path,
        top_path: Path,
        container_mtime: Optional[float] = None,
        convert: bool = True,
    ) -> DirResult:
        """Recursively optimize a directory."""
        dir_result = DirResult(dir_path, [])
        for path in dir_path.iterdir():
            result = self.walk_file(path, top_path, container_mtime, convert=convert)
            dir_result.results.append(result)

        return dir_result

    def _walk_container_dir(
        self, top_path: Path, handler: ContainerHandler
    ) -> Union[ContainerResult, ApplyResult]:
        """Optimize a container."""
        result: Union[ContainerResult, ApplyResult]
        try:
            container_mtime = handler.original_path.stat().st_mtime
            dir_result = self.walk_dir(
                handler.tmp_container_dir,
                top_path,
                container_mtime,
                convert=handler.CONVERT,
            )
            result = ContainerResult(handler.final_path, dir_result.results, handler)
        except Exception as exc:
            args = tuple([exc])
            result = self._pool.apply_async(handler.error, args=args)
        return result

    def _report_totals(self, totals: Totals) -> None:
        """Report the total number and percent of bytes saved."""
        if totals.bytes_in:
            bytes_saved = totals.bytes_in - totals.bytes_out
            percent_bytes_saved = bytes_saved / totals.bytes_in * 100
            msg = ""
            if self._config.test:
                if percent_bytes_saved > 0:
                    msg += "Could save"
                elif percent_bytes_saved == 0:
                    msg += "Could even out for"
                else:
                    msg += "Could lose"
            else:
                if percent_bytes_saved > 0:
                    msg += "Saved"
                elif percent_bytes_saved == 0:
                    msg += "Evened out"
                else:
                    msg = "Lost"
            msg += " a total of {} or {:.{prec}f}%".format(
                naturalsize(bytes_saved), percent_bytes_saved, prec=2
            )
            if self._config.verbose:
                print(msg)
                if self._config.test:
                    print("Test run did not change any files.")

        else:
            if self._config.verbose:
                print("Didn't optimize any files.")

        if totals.errors:
            cprint("Errors with the following files:", "yellow")
            for rs in totals.errors:
                rs.report(self._config.test, "yellow")

    def _handle_queue_item_dir(self, top_path: Path, dir_result: DirResult):
        """Reverse the results tree to handle directories bottom up."""
        queue = self._queues[top_path]
        for dir_member_result in dir_result.results:
            if isinstance(dir_member_result, DirResult):
                self._handle_queue_item_dir(top_path, dir_member_result)
            else:
                queue.put(dir_member_result)
        if isinstance(dir_result, ContainerResult):
            task = CompleteContainerTask(dir_result.handler)
        else:
            task = CompleteDirTask(dir_result.path)
        queue.put(task)

    def _handle_queue_complete_task(self, top_path: Path, task: CompleteTask):
        if isinstance(task, CompleteDirTask):
            if self._config.timestamps:
                # Compact timestamps after every directory completes
                timestamps = self._timestamps[top_path]
                timestamps.set(task.path, compact=True)
        elif isinstance(task, CompleteContainerTask):
            # Repack inline, not in pool, to complete directories immediately
            repack_result = task.handler.repack()
            self._queues[top_path].put(repack_result)
        else:
            cprint(f"WARNING: Unhandled Complete task {task}", "yellow")

    def _handle_queue_item(
        self,
        top_path: Path,
        totals: Totals,
    ):
        queue: SimpleQueue = self._queues[top_path]
        item = queue.get()
        if isinstance(item, ApplyResult):
            # unpack apply results inline to preserve directory order
            item = item.get()

        if item is None:
            pass
        elif isinstance(item, ReportStats):
            if item.error:
                cprint(item.error, "yellow")
                totals.errors.append(item)
            else:
                if self._config.timestamps:
                    timestamps = self._timestamps[top_path]
                    timestamps.set(item.path)
                totals.bytes_in += item.bytes_in
                totals.bytes_out += item.bytes_out
        elif isinstance(item, DirResult):
            self._handle_queue_item_dir(top_path, item)
        elif isinstance(item, CompleteTask):
            self._handle_queue_complete_task(top_path, item)
        else:
            cprint(f"Unhandled queue item {item}", "yellow")

    def _convert_message(
        self, convert_from_formats: frozenset[str], convert_handler: Type[Handler]
    ):
        convert_from = ", ".join(
            sorted(convert_from_formats & frozenset(self._config.formats))
        )
        convert_to = convert_handler.OUTPUT_FORMAT
        cprint(f"Converting {convert_from} to {convert_to}", "cyan")

    def _get_covertable_formats(
        self, convertable_formats: frozenset[str]
    ) -> frozenset[str]:
        formats = set(convertable_formats)
        if TIFF_FORMAT in self._config.formats:
            formats.add(TIFF_FORMAT)
        return frozenset(formats)

    def _init_run(self):
        """Init Run."""
        # Validate top_paths
        if not self._top_paths:
            raise ValueError("No paths to optimize.")
        for path in self._top_paths:
            if not path.exists():
                raise ValueError(f"Path does not exist: {path}")

        # Tell the user what we're doing
        if self._config.verbose:
            print("Optimizing formats:", *sorted(self._config.formats))
            if self._config.convert_to:
                if WebP.OUTPUT_FORMAT in self._config.convert_to:
                    formats = self._get_covertable_formats(WEBP_CONVERTABLE_FORMATS)
                    self._convert_message(formats, WebP)
                elif Png.OUTPUT_FORMAT in self._config.convert_to:
                    formats = self._get_covertable_formats(PNG_CONVERTABLE_FORMATS)
                    self._convert_message(formats, Png)
                if Zip.OUTPUT_FORMAT in self._config.convert_to:
                    self._convert_message(frozenset([Zip.INPUT_FORMAT_RAR]), Zip)
                if CBZ.OUTPUT_FORMAT in self._config.convert_to:
                    self._convert_message(frozenset([CBZ.INPUT_FORMAT_RAR]), CBZ)
            if self._config.after is not None:
                print("Optimizing after", time.ctime(self._config.after))

        # Init timestamps
        if self._config.timestamps:
            self._timestamps = Treestamps.path_to_treestamps_map_factory(
                self._top_paths,
                PROGRAM_NAME,
                self._config.verbose,
                self._config,
                TIMESTAMPS_CONFIG_KEYS,
            )
            for timestamps in self._timestamps.values():
                OldTimestamps(timestamps).import_old_timestamps()

    def run(self) -> bool:
        """Optimize all configured files."""
        try:
            self._init_run()
        except Exception as exc:
            cprint(str(exc), "red")
            return False

        # Start each queue
        totals = Totals()
        for top_path in self._top_paths:
            dirpath = Treestamps.dirpath(top_path)
            result = self.walk_file(top_path, dirpath)
            if dirpath not in self._queues:
                self._queues[dirpath] = SimpleQueue()
            queue = self._queues[dirpath]
            self._queues[dirpath].put(result)

        # Process each queue
        for top_path, queue in self._queues.items():
            while not queue.empty():
                self._handle_queue_item(top_path, totals)

        # Shut down multiprocessing
        self._pool.close()
        self._pool.join()

        if self._config.timestamps:
            for timestamps in self._timestamps.values():
                timestamps.dump()
        # Finish by reporting totals
        self._report_totals(totals)
        return True



"""FileType abstract class for image and container formats."""
import shutil
import subprocess

from abc import ABC
from dataclasses import dataclass
from pathlib import Path
from typing import Optional

from confuse.templates import AttrDict
from termcolor import cprint

from picopt import PROGRAM_NAME
from picopt.stats import ReportStats


@dataclass(eq=True, frozen=True)
class Format:
    """A file format, with image attributes."""

    format: str
    lossless: bool = True
    animated: bool = False


@dataclass(eq=True, frozen=True)
class Metadata:
    """Image metadata class."""

    exif: bytes = b""
    icc_profile: str = ""
    n_frames: int = 1


class Handler(ABC):
    """FileType superclass for image and container formats."""

    BEST_ONLY: bool = True
    OUTPUT_FORMAT: str = "unimplemented"
    OUTPUT_FORMAT_OBJ: Format = Format(OUTPUT_FORMAT, False, False)
    INTERNAL: str = "python_internal"
    PROGRAMS: dict[str, Optional[str]] = {}
    WORKING_SUFFIX: str = f"{PROGRAM_NAME}__tmp"

    @classmethod
    def init_programs(cls, programs: tuple[str, ...]) -> dict[str, Optional[str]]:
        """Initialize the PROGRAM map."""
        program_dict = {}
        for program in programs:
            bin_path = None
            if not program.startswith("pil2") and program != cls.INTERNAL:
                bin_path = shutil.which(program)
            program_dict[program] = bin_path
        return program_dict

    @staticmethod
    def run_ext(args: tuple[str, ...]) -> None:
        """Run EXTERNAL program."""
        try:
            if not args[0]:
                raise ValueError(f"{args}")
            subprocess.run(args, check=True)
        except subprocess.CalledProcessError as exc:
            cprint(str(exc), "red")
            cprint(exc.cmd, "red")
            cprint(f"retcode: {exc.returncode}", "red")
            cprint(str(exc.output), "red")
            raise

    @classmethod
    def native_input_format_objs(cls) -> set[Format]:
        """Return input formats handled without conversion."""
        return set([cls.OUTPUT_FORMAT_OBJ])

    @classmethod
    def _output_suffix(cls) -> str:
        """Return the suffix without a leading dot."""
        return cls.OUTPUT_FORMAT.lower()

    @classmethod
    def output_suffix(cls) -> str:
        """Generate the output suffix for the handler."""
        return "." + cls._output_suffix()

    @classmethod
    def is_handler_available(
        cls,
        convert_handlers: dict,
        available_programs: set,
        format_obj: Format,
    ):
        """Can this handler run with available programs."""
        handled_format_objs = cls.native_input_format_objs() | convert_handlers.get(
            cls, set()
        )
        return format_obj in handled_format_objs and bool(
            available_programs & set(cls.PROGRAMS.keys())
        )

    def __init__(
        self,
        config: AttrDict,
        original_path: Path,
        input_format_obj: Format,
        metadata: Metadata,
    ):
        """Initialize handler."""
        self.config: AttrDict = config
        self.original_path: Path = original_path
        self.working_paths: set[Path] = set()
        self.final_path: Path = self.original_path.with_suffix(self.output_suffix())
        self.input_format_obj: Format = input_format_obj
        self.metadata = metadata

    def get_working_path(self, identifier: str = "") -> Path:
        """Return a working path with a custom suffix."""
        suffixes = [self.original_path.suffix, self.WORKING_SUFFIX]
        if identifier:
            suffixes += [identifier]
        suffixes += [self._output_suffix()]
        suffix = ".".join(suffixes)

        return self.original_path.with_suffix(suffix)

    def _cleanup_after_optimize_aux(self, last_working_path: Path) -> tuple[int, int]:
        """Replace old file with better one or discard new wasteful file."""
        bytes_in = 0
        bytes_out = 0
        try:
            bytes_in = self.original_path.stat().st_size
            bytes_out = last_working_path.stat().st_size
            if not self.config.test and (
                (bytes_out > 0) and ((bytes_out < bytes_in) or self.config.bigger)
            ):
                last_working_path.replace(self.final_path)
                if self.final_path != self.original_path:
                    self.working_paths.add(self.original_path)
            else:
                self.working_paths.add(last_working_path)
                bytes_out = bytes_in
            if self.final_path in self.working_paths:
                self.working_paths.remove(self.final_path)
            for working_path in self.working_paths:
                working_path.unlink(missing_ok=True)
        except OSError as exc:
            cprint(f"ERROR: cleanup_after_optimize: {exc}", "red")
            raise

        return (bytes_in, bytes_out)

    def cleanup_after_optimize(self, last_working_path: Path) -> tuple[int, int]:
        """
        Replace old file with better one or discard new wasteful file.

        And report results using the stats module.
        """
        return self._cleanup_after_optimize_aux(last_working_path)

    def error(self, exc: Exception) -> ReportStats:
        """Return an error result."""
        report_stats = ReportStats(self.original_path, error=str(exc))
        if self.config.verbose:
            report_stats.report(self.config.test)
        return report_stats



#!/usr/bin/env python3
"""
Get the bit depth of a png image manually.

https://www.w3.org/TR/PNG-Chunks.html
This should be a part of Pillow
"""
from pathlib import Path
from typing import Optional

from termcolor import cprint

from picopt.pillow.unpack import compare_header, unpack


PNG_HEADER = (b"\x89", b"P", b"N", b"G", b"\r", b"\n", b"\x1a", b"\n")


def png_bit_depth(path: Path) -> Optional[int]:
    """If a file is a png, get the bit depth from the standard position."""
    try:
        with path.open("rb") as img:
            compare_header(img, 0, PNG_HEADER)

            img.seek(24)  # bit depth offset
            depth = unpack("b", 1, img)[0]
        return int(depth)
    except ValueError:
        cprint(f"WARNING: {path} is not a png!", "yellow")
        return None


def main() -> None:
    """Stand alone cli tool for getting png bit depth."""
    import sys

    bit_depth = png_bit_depth(Path(sys.argv[1]))
    print(bit_depth)


if __name__ == "__main__":
    main()



"""WebP Animated images are treated like containers."""
from pathlib import Path
from typing import Optional

from PIL import Image, ImageSequence

from picopt.handlers.container import ContainerHandler
from picopt.handlers.handler import Format
from picopt.handlers.image import TIFF_ANIMATED_FORMAT_OBJ
from picopt.handlers.webp import WebP


class WebPAnimated(ContainerHandler):
    """Animated WebP container."""

    OUTPUT_FORMAT: str = WebP.OUTPUT_FORMAT
    OUTPUT_FORMAT_OBJ = Format(OUTPUT_FORMAT, False, True)
    PROGRAMS = ContainerHandler.init_programs(("webpmux", "img2webp"))
    _OPEN_WITH_PIL_FORMAT_OBJS = set([TIFF_ANIMATED_FORMAT_OBJ])
    _IMG2WEBP_ARGS_PREFIX = (PROGRAMS["img2webp"], "-min_size")
    _WEBPMUX_ARGS_PREFIX = (PROGRAMS["webpmux"], "-get", "frame")

    @classmethod
    def identify_format(cls, _path: Path) -> Optional[Format]:
        """Return the format if this handler can handle this path."""
        raise NotImplementedError()

    def _get_frame_path(self, frame_index: int) -> Path:
        """Return a frame path for an index."""
        return self.tmp_container_dir / f"frame-{frame_index:08d}.webp"

    def unpack_into(self) -> None:
        """Unpack webp into temp dir."""
        if self.input_format_obj in self._OPEN_WITH_PIL_FORMAT_OBJS:
            with Image.open(self.original_path) as image:
                frame_index = 0
                for frame in ImageSequence.Iterator(image):
                    frame_path = self._get_frame_path(frame_index)
                    frame.save(
                        frame_path,
                        self.OUTPUT_FORMAT,
                        lossless=True,
                        quality=100,
                        method=0,
                    )
                    frame_index += 1
            image.close()
        else:
            for frame_index in range(0, self.metadata.n_frames):
                frame_path = self._get_frame_path(frame_index)
                args = [
                    *self._WEBPMUX_ARGS_PREFIX,
                    str(frame_index),
                    str(self.original_path),
                    "-o",
                    str(frame_path),
                ]
                self.run_ext(tuple(args))

    def _prepare_metadata(self, data: Optional[bytes], working_path: Path, md_arg: str):
        """Prepare a metadata file and args for webpmux."""
        if not data:
            return []
        md_path = working_path.with_suffix("." + md_arg)
        with md_path.open("wb") as md_file:
            md_file.write(data)
        self.working_paths.add(md_path)
        return ["-set", md_arg, str(md_path)]

    def _set_metadata(self, working_path):
        """Set the exif data on the rebuilt image."""
        if not self.metadata.exif or self.metadata.icc_profile:
            return

        args = ["webpmux"]
        # dump exif
        if self.metadata.exif:
            args += self._prepare_metadata(self.metadata.exif, working_path, "exif")

        if self.metadata.icc_profile:
            args += self._prepare_metadata(
                self.metadata.icc_profile.encode(), working_path, "icc"
            )

        # move working file
        container_exif_working_path = self.get_working_path("exif")
        working_path.replace(container_exif_working_path)
        self.working_paths.add(container_exif_working_path)

        # run exif set
        args += [
            str(container_exif_working_path),
            "-o",
            str(working_path),
        ]
        self.run_ext(tuple(args))

    def pack_into(self, working_path: Path) -> None:
        """Remux the optimized frames into an animated webp."""
        frames = sorted([str(path) for path in self.tmp_container_dir.iterdir()])
        args = [
            *self._IMG2WEBP_ARGS_PREFIX,
            *frames,
            "-o",
            str(working_path),
        ]
        self.run_ext(tuple(args))
        if self.config.keep_metadata:
            self._set_metadata(working_path)



"""Test comic format."""
import platform
import shutil

from picopt import PROGRAM_NAME, cli
from tests import IMAGES_DIR, get_test_dir


__all__ = ()
TMP_ROOT = get_test_dir()
FNS = {
    "07themecamplist.pdf": (93676, 93676, ("pdf", 93676), ("pdf", 93676)),
    "test_animated_gif.gif": (16383, 16358, ("gif", 16358), ("webp", 11846)),
    "test_animated_webp.webp": (13610, 11854, ("webp", 11854), ("webp", 11854)),
    "test_png.png": (7967, 4379, ("png", 4379), ("webp", 3870)),
    "test_pre-optimized_jpg.jpg": (
        22664,
        22664,
        ("jpg", 22664),
        ("jpg", 22664),
    ),
    "test_txt.txt": (6, 6, ("txt", 6), ("txt", 6)),
    "test_webp_lossless.webp": (5334, 3870, ("webp", 3870), ("webp", 3870)),
    "test_webp_lossless_pre-optimized.webp": (
        8914,
        8914,
        ("webp", 8914),
        ("webp", 8914),
    ),
    "mri.tif": (230578, 230578, ("tif", 230578), ("webp", 128416)),
}
if platform.system() == "Darwin":
    FNS.update(
        {
            "test_bmp.bmp": (141430, 141430, ("png", 67215), ("webp", 47524)),
            "test_gif.gif": (138952, 138944, ("png", 112467), ("webp", 26504)),
            "test_jpg.jpg": (97373, 87913, ("jpg", 87913), ("jpg", 87913)),
            "test_png_16rgba.png": (3435, 2870, ("png", 2870), ("webp", 1142)),
            "test_pnm.pnm": (27661, 27661, ("png", 15510), ("webp", 12808)),
            "test_pre-optimized_png.png": (
                256572,
                256572,
                ("png", 256572),
                ("webp", 197680),
            ),
            "test_webp_lossy.webp": (2764, 1760, ("webp", 1760), ("webp", 1760)),
            "test_webp_lossy_pre-optimized.webp": (
                1514,
                1514,
                ("webp", 1514),
                ("webp", 1514),
            ),
            "eight.tif": (59640, 59640, ("png", 30564), ("webp", 24982)),
        }
    )
else:
    FNS.update(
        {
            "test_bmp.bmp": (141430, 141430, ("png", 67215), ("webp", 47594)),
            "test_gif.gif": (138952, 138944, ("png", 112467), ("webp", 26148)),
            "test_jpg.jpg": (97373, 87922, ("jpg", 87922), ("jpg", 87922)),
            "test_png_16rgba.png": (3435, 2870, ("png", 2870), ("webp", 684)),
            "test_pnm.pnm": (27661, 27661, ("png", 15510), ("webp", 12816)),
            "test_pre-optimized_png.png": (
                256572,
                256572,
                ("png", 256572),
                ("webp", 197750),
            ),
            "test_webp_lossy.webp": (2764, 1704, ("webp", 1704), ("webp", 1704)),
            "test_webp_lossy_pre-optimized.webp": (  # XXX SHOULD NOT CONVERT MORE?
                1514,
                1496,
                ("webp", 1496),
                ("webp", 1496),
            ),
            "eight.tif": (59640, 59640, ("png", 30564), ("webp", 26642)),
        }
    )


class TestImagesDir:
    def setup_method(self) -> None:
        self.teardown_method()
        shutil.copytree(IMAGES_DIR, TMP_ROOT)
        for name, sizes in FNS.items():
            path = TMP_ROOT / name
            assert path.stat().st_size == sizes[0]

    def teardown_method(self) -> None:
        if TMP_ROOT.exists():
            shutil.rmtree(TMP_ROOT)

    def test_no_convert(self) -> None:
        args = (PROGRAM_NAME, "-rvv", str(TMP_ROOT))
        res = cli.main(args)
        assert res
        for name, sizes in FNS.items():
            path = TMP_ROOT / name
            assert path.stat().st_size == sizes[1]

    def test_convert_to_png(self) -> None:
        args = (PROGRAM_NAME, "-rvvx", "TIFF", "-c", "PNG", str(TMP_ROOT))
        res = cli.main(args)
        assert res
        for name, sizes in FNS.items():
            path = (TMP_ROOT / name).with_suffix("." + sizes[2][0])
            assert path.stat().st_size == sizes[2][1]

    def test_convert_to_webp(self) -> None:
        args = (PROGRAM_NAME, "-rvvx", "TIFF", "-c", "WEBP", str(TMP_ROOT))
        res = cli.main(args)
        assert res
        for name, sizes in FNS.items():
            path = (TMP_ROOT / name).with_suffix("." + sizes[3][0])
            assert path.stat().st_size == sizes[3][1]



"""WebP format."""
from abc import ABC
from pathlib import Path
from typing import Any, Optional

from PIL import Image
from PIL.WebPImagePlugin import WebPImageFile

from picopt.handlers.handler import Format
from picopt.handlers.image import CONVERTABLE_FORMAT_OBJS, TIFF_FORMAT_OBJ, ImageHandler
from picopt.handlers.png import Png


class WebPBase(ImageHandler, ABC):
    """Base for handlers that use WebP utility commands."""

    PIL2WEBP_KWARGS: dict[str, Any] = {"lossless": True, "quality": 100, "method": 6}
    PIL2_ARGS: dict[str, Any] = PIL2WEBP_KWARGS

    def get_metadata_args(self) -> list[str]:
        """Get webp utility metadata args."""
        args = ["-metadata"]
        if self.config.keep_metadata:
            args += ["all"]
        else:
            args += ["none"]
        return args

    def pil2webp(self, old_path: Path, new_path: Path) -> Path:
        """Pillow webp optimization."""
        return self.pil2native(old_path, new_path)


class WebP(WebPBase, ABC):
    """WebP format class."""

    OUTPUT_FORMAT = WebPImageFile.format
    PROGRAMS: dict[str, Optional[str]] = WebPBase.init_programs(("cwebp",))
    ARGS_PREFIX = [
        PROGRAMS["cwebp"],
        "-mt",
        "-quiet",
        "-sharp_yuv",
        "-af",
        "-alpha_filter",
        "best",
    ]

    def cwebp(self, old_path: Path, new_path: Path) -> Path:
        """Optimize using cwebp."""
        args = tuple(
            self.ARGS_PREFIX
            + self.get_metadata_args()
            + [str(old_path), "-o", str(new_path)]
        )
        self.run_ext(args)
        return new_path


class WebPLossless(WebP):
    """Handle lossless webp images and images that convert to lossless webp."""

    BEST_ONLY: bool = False
    OUTPUT_FORMAT_OBJ = Format(WebP.OUTPUT_FORMAT, True, False)
    PREFERRED_PROGRAM: str = "cwebp"
    PROGRAMS: dict[str, Optional[str]] = {
        "pil2png": None,
        **WebP.PROGRAMS,
        "pil2webp": None,
    }
    ARGS_PREFIX = WebP.ARGS_PREFIX + [
        "-lossless",
        "-z",
        "9",
    ]
    _PIL2PNG_FORMATS = CONVERTABLE_FORMAT_OBJS | set([TIFF_FORMAT_OBJ])

    def pil2png(self, old_path: Path, new_path: Path) -> Path:
        """Internally convert unhandled formats to uncompressed png for cwebp."""
        if (
            self.input_format_obj in self._PIL2PNG_FORMATS
            and self.PREFERRED_PROGRAM in self.config._available_programs
        ):
            new_path = new_path.with_suffix(Png.output_suffix())
            with Image.open(old_path) as image:
                image.save(
                    new_path,
                    Png.OUTPUT_FORMAT,
                    compress_level=0,
                    exif=self.metadata.exif,
                    icc_profile=self.metadata.icc_profile,
                )
            image.close()
            self.input_format_obj = Png.OUTPUT_FORMAT_OBJ
        else:
            new_path = old_path
        return new_path


class WebPLossy(WebP):
    """Handle lossy webp images."""

    OUTPUT_FORMAT_OBJ = Format(WebP.OUTPUT_FORMAT, False, False)
    ARGS_PREFIX = WebP.ARGS_PREFIX + [
        "-m",
        "6",
        "-pass",
        "10",
    ]


class Gif2WebP(WebPBase):
    """
    Animated WebP format class.

    There are no easy animated WebP optimization tools. So this only
    converts animated gifs.
    """

    OUTPUT_FORMAT = WebP.OUTPUT_FORMAT
    OUTPUT_FORMAT_OBJ = Format(WebP.OUTPUT_FORMAT, True, True)
    PIL2WEBP_KWARGS: dict[str, Any] = {
        **WebPLossless.PIL2WEBP_KWARGS,
        "minimize_size": True,
    }
    PREFERRED_PROGRAM = "gif2webp"
    PROGRAMS: dict[str, Optional[str]] = WebPBase.init_programs(
        ("gif2webp", "pil2webp")
    )
    _ARGS_PREFIX = [
        PROGRAMS["gif2webp"],
        "-mixed",
        "-min_size",
        "-m",
        "6",
        "-mt",
    ]

    @classmethod
    def native_input_format_objs(cls):
        """No native formats."""
        return set()

    def gif2webp(self, old_path: Path, new_path: Path) -> Path:
        """Convert animated gif to animated webp."""
        args = tuple(
            self._ARGS_PREFIX
            + self.get_metadata_args()
            + [str(old_path), "-o", str(new_path)]
        )
        self.run_ext(args)
        return new_path



"""Unpack items from a file descriptor."""
import struct

from typing import BinaryIO, Union


def unpack(
    fmt_type: str, length: int, file_desc: BinaryIO
) -> Union[tuple[bytes, ...], tuple[int]]:
    """Unpack information from a file according to format string & length."""
    return struct.unpack(fmt_type * length, file_desc.read(length))


def compare_header(img: BinaryIO, seek: int, byte_array: tuple[bytes, ...]) -> None:
    """Seek to a spot in a binary file and compare a byte array."""
    img.seek(seek)
    compare = unpack("c", len(byte_array), img)
    if compare != byte_array:
        raise ValueError(f"{compare} != {byte_array}")



"""Test comic format."""
import shutil

from platform import system
from zipfile import ZipFile

from picopt import PROGRAM_NAME, cli
from tests import CONTAINER_DIR, get_test_dir


__all__ = ()
TMP_ROOT = get_test_dir()
SRC_CBZ = CONTAINER_DIR / "test_cbz.cbz"

if system() == "Darwin":
    FNS = {
        "test_cbz.cbz": (93408, 84493, ("cbz", 93408)),
        "test_cbr.cbr": (93725, 93725, ("cbz", 84506)),
        "test_rar.rar": (93675, 93675, ("zip", 84493)),
        "test_zip.zip": (2974, 1917, ("zip", 2974)),
        "igp-twss.epub": (292448, 281248, ("epub", 292448)),
    }
else:
    FNS = {
        "test_cbz.cbz": (93408, 84481, ("cbz", 93408)),
        "test_cbr.cbr": (93725, 93725, ("cbz", 84494)),
        "test_rar.rar": (93675, 93675, ("zip", 84481)),
        "test_zip.zip": (2974, 1861, ("zip", 2974)),
        "igp-twss.epub": (292448, 280700, ("epub", 292448)),
    }


class TestContainersDir:
    def setup_method(self) -> None:
        self.teardown_method()
        shutil.copytree(CONTAINER_DIR, TMP_ROOT)
        for name, sizes in FNS.items():
            path = TMP_ROOT / name
            assert path.stat().st_size == sizes[0]

    def teardown_method(self) -> None:
        if TMP_ROOT.exists():
            shutil.rmtree(TMP_ROOT)

    def test_containers_noop(self) -> None:
        args = (PROGRAM_NAME, "-r", str(TMP_ROOT))
        res = cli.main(args)
        assert res
        for name, sizes in FNS.items():
            if name == "igp-twss.epub":
                path = TMP_ROOT / name
                with ZipFile(path, "r") as zf:
                    namelist = zf.namelist()
                assert "OPS/test_bmp.bmp" in namelist
            path = TMP_ROOT / name
            assert path.stat().st_size == sizes[0]

    def test_containers_no_convert(self) -> None:
        args = (PROGRAM_NAME, "-rx", "CBZ,ZIP,EPUB", "-c WEBP", str(TMP_ROOT))
        res = cli.main(args)
        assert res
        for name, sizes in FNS.items():
            path = TMP_ROOT / name
            if name == "igp-twss.epub":
                with ZipFile(path, "r") as zf:
                    namelist = zf.namelist()
                assert "OPS/test_bmp.bmp" in namelist
            assert path.stat().st_size == sizes[1]

    def test_containers_convert_to_zip(self) -> None:
        args = (PROGRAM_NAME, "-rc", "ZIP,CBZ", str(TMP_ROOT))
        res = cli.main(args)
        assert res
        for name, sizes in FNS.items():
            path = (TMP_ROOT / name).with_suffix("." + sizes[2][0])
            assert path.stat().st_size == sizes[2][1]



"""Gif format."""
import shutil

from pathlib import Path
from typing import Optional

from PIL import Image
from PIL.GifImagePlugin import GifImageFile

from picopt.handlers.handler import Format
from picopt.handlers.image import ImageHandler


class Gif(ImageHandler):
    """GIF handler."""

    OUTPUT_FORMAT = GifImageFile.format
    OUTPUT_FORMAT_OBJ = Format(OUTPUT_FORMAT, True, False)
    PROGRAMS: dict[str, Optional[str]] = ImageHandler.init_programs(
        ("gifsicle", "pil2gif")
    )
    _ARGS_PREFIX = [
        PROGRAMS["gifsicle"],
        "--optimize=3",
        "--batch",
    ]

    def gifsicle(self, old_path: Path, new_path: Path) -> Path:
        """Return gifsicle args."""
        if not self._ARGS_PREFIX[0]:
            return old_path

        shutil.copy2(old_path, new_path)
        args = tuple(self._ARGS_PREFIX + [str(new_path)])
        self.run_ext(args)
        return new_path

    def pil2gif(self, old_path: Path, new_path: Path) -> Path:
        """Pillow gif optimization."""
        with Image.open(old_path) as image:
            image.save(new_path, self.OUTPUT_FORMAT, optimize=True, save_all=True)
        image.close()  # for animated images
        return new_path


class AnimatedGif(Gif):
    """Animated GIF handler."""

    OUTPUT_FORMAT_OBJ = Format(Gif.OUTPUT_FORMAT, True, True)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['picopt',
 'picopt.handlers',
 'picopt.pillow',
 'tests',
 'tests.integration',
 'tests.unit']

package_data = \
{'': ['*'],
 'tests': ['test_files/containers/*',
           'test_files/images/*',
           'test_files/invalid/*']}

install_requires = \
['Pillow>=9.0,<10.0',
 'confuse>=1.7.0,<2.0.0',
 'humanize>=4.0.0,<5.0.0',
 'python-dateutil>=2.8,<3.0',
 'rarfile>=4.0,<5.0',
 'ruamel.yaml>=0.17.16,<0.18.0',
 'termcolor>=1.1.0,<2.0.0',
 'treestamps>=0.1.2,<0.2.0']

entry_points = \
{'console_scripts': ['picopt = picopt.cli:main']}

setup_kwargs = {
    'name': 'picopt',
    'version': '3.0.3',
    'description': 'A multi format lossless image optimizer that uses external tools',
    'long_description': '# picopt\n\nA multi-format, recursive, multiprocessor aware, command line lossless image optimizer utility that uses external tools to do the optimizing.\n\nPicopt depends on Python [PIL](http://www.pythonware.com/products/pil/) to identify files and Python [rarfile](https://pypi.python.org/pypi/rarfile) to open CBRs.\n\nPicopt will optionally drop hidden timestamps at the root of your image directories to avoid reoptimizing images picopt has already optimized.\n\nThe actual image optimization is best accomplished by external programs.\n\n## <a name="philosophy">Conversion Philosophy</a>\n\n### Lossy Images\n\nJPEG & Lossy WebP images are likely the best and most practical lossy image formats. Converting lossy images rarely makes sense and so picopt only optimizes them in their current format.\n\n### Lossless Images\n\nLossless WebP images are smaller than PNG, much smaller than GIF and, of course, a great deal smaller thein uncompressed bitmaps like BMP. As such the best practice is probably to convert all lossless images to WebP Lossless as now all major browsers support it. The only downside is that decoding WebP Lossless takes on average 50% more CPU than PNG.\n\n### Sequenced Images\n\nSequenced Images, like animated GIFs and WebP, most of the time, should be converted to a compressed video format like HEVC or VP9. There are several situations where this is impractical and so Animated WebP is now a good substitute.\n\n### Conversion\n\nBy default picopt does not convert images between formats. You must turn on conversion to PNG or WebP explicitly.\n\n## <a name="formats">Formats</a>\n\n- By default picopt will optimize GIF, JPEG, PNG and WEBP images.\n- Picopt can optionally optimize ZIP, ePub, and CBZ containers.\n- Picopt can be told to convert lossless images such as BPM, PPM, GIF, TIFF into PNG, and all of the mentioned lossless formats into WebP.\n- Picopt can convert Animated GIFS into Animated WebP files.\n- Picopt can convert RAR files into Zipfiles and CBR files into CBZ files.\n\n## <a name="programs">External Programs</a>\n\nPicopt will perform some minor optimization on most formats natively without using external programs, but this is not very good compared to the optimizations external programs can provide.\n\n### JPEG\n\nTo optimize JPEG images. Picopt needs one of [mozjpeg](https://github.com/mozilla/mozjpeg) or [jpegtran](http://jpegclub.org/jpegtran/) on the path. in order of preference.\n\n### PNG\n\nTo optimize PNG images or convert other lossless formats to PNG picopt requires either [optipng](http://optipng.sourceforge.net/) or [pngout](http://advsys.net/ken/utils.htm) be on the path. Optipng provides the most advantage, but best results will be had by using pngout as well.\n\n### Animated GIF\n\nAnimated GIFs are optimized with [gifsicle](http://www.lcdf.org/gifsicle/) if it is available.\n\n### WebP\n\nWebP lossless & lossy formats are optimized with [cwebp](https://developers.google.com/speed/webp/docs/cwebp).\n\n### EPub\n\nEPub Books are zip files that often contain images and picopt unpacks and repacks this format natively. Images within the epub are handled by other programs. EPub optimization is not turned on by default.\nEPub contents are never converted to other formats because it would break internal references to them.\n\n### CBZ & CBR\n\nPicopt uncompresses, optimizes and rezips [comic book archive files](https://en.wikipedia.org/wiki/Comic_book_archive). Be aware that CBR rar archives may only be rezipped into CBZs instead of CBR. Comic book archive optimization is not turned on by default to prevent surprises.\n\n## <a name="install">Install</a>\n\n### System Dependencies\n\npicopt requires several external system dependencies to run. We must install these first\n\n#### macOS\n\n    brew install webp mozjpeg optipng jonof/kenutils/pngout gifsicle\n\nln -s $(brew --prefix)/opt/mozjpeg/bin/jpegtran /usr/local/bin/mozjpeg\n\nUnfortunately hombrew\'s `webp` formula does not yet install the gif2webp tool that picopt uses for converting animated gifs to animated webps.\nYou may manually download it and put it in your path at [Google\'s WebP developer website](https://developers.google.com/speed/webp/download)\n\n#### Debian / Ubuntu\n\n    apt-get install optipng gifsicle python-imaging webp\n\nif you don\'t want to install mozjpeg using the instructions below then use jpegtran:\n\n    apt-get install libjpeg-progs\n\n#### Redhat / Fedora\n\n    yum install optipng gifsicle python-imaging libwebp-tools\n\nif you don\'t want to install mozjpeg using the instructions below then use jpegtran:\n\n    yum install libjpeg-progs\n\n#### MozJPEG\n\nmozjpeg offers better compression than libjpeg-progs jpegtran. It may or\nmay not be packaged for your \\*nix, but even when it is, picopt requires that its separately compiled version of jpegtran be symlinked to \'mozjpeg\' somewhere in the path.\n\nInstructions for installing on macOS are given above.\nSome near recent binaries for Windows and Debian x86 [can be found here](https://mozjpeg.codelove.de/binaries.html).\nMost Linux distributions still require a more manual install as elucidated here on [Casey Hoffer\'s blog](https://www.caseyhofford.com/2019/05/01/improved-image-compression-install-mozjpeg-on-ubuntu-server/)\n\n#### pngout\n\npngout is a useful compression to use after optipng. It is not packaged for linux, but you may find the latest binary version [on JonoF\'s site](http://www.jonof.id.au/kenutils). Picopt looks for the binary to be called `pngout`\n\n### Picopt python package\n\n    pip install picopt\n\n## <a name="usage">Usage</a>\n\nOptimize all JPEG files in a directory:\n\n    picopt *.jpg\n\nOptimize all files and recurse directories:\n\n    picopt -r *\n\nOptimize files, recurse directories, also optimize ePub & CBZ containers, convert lossless images into WEBP, convert CBR into CBZ.\n\n    picopt -rx EPUB -c WEBP,CBZ *\n\nOptimize files and recurse directories AND optimize comic book archives:\n\n    picopt -rx CBZ *\n\nOptimize all files, but only JPEG format files:\n\n    picopt -f JPEG *\n\nOptimize files and containers, but not JPEGS:\n\n    picopt -f GIF,PNG,WEBP,ZIP,CBZ,EPUB *\n\nOptimize files, but not animated gifs:\n\n    picopt -f PNG,WEBP,ZIP,CBZ,EPUB *\n\nJust list files picopt.py would try to optimize:\n\n    picopt -L *\n\nOptimize everything in my iPhoto library, but only after the last time I did this, skipping symlinks to avoid duplicate work. Also drop a timestamp file so I don\'t have to remember the last time I did this:\n\n    picopt -rSt -D \'2013 June 1 14:00\' \'Pictures/iPhoto Library\'\n\n## <a name="package">Packages</a>\n\n- [PyPI](https://pypi.python.org/pypi/picopt/)\n- [Arch Linux](https://aur.archlinux.org/packages/picopt/)\n\n## <a name="alternatives">Alternatives</a>\n\n[imagemin](https://github.com/imagemin/imagemin-cli) looks to be an all in one cli and gui solution with bundled libraries, so no awkward dependencies.\n[Imageoptim](http://imageoptim.com/) is an all-in-one OS X GUI image optimizer. Imageoptim command line usage is possible with [an external program](https://code.google.com/p/imageoptim/issues/detail?can=2&start=0&num=100&q=&colspec=ID%20Type%20Status%20Priority%20Milestone%20Owner%20Summary%20Stars&groupby=&sort=&id=39).\n',
    'author': 'AJ Slater',
    'author_email': 'aj@slater.net',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/ajslater/picopt',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



"""import old picopt timestamps to new ones."""
import os

from pathlib import Path

from treestamps import Treestamps

from picopt import PROGRAM_NAME


_OLD_TIMESTAMPS_NAME = f".{PROGRAM_NAME}_timestamp"


class OldTimestamps:
    """Old timestamps importer."""

    def _add_old_timestamp(self, old_timestamp_path: Path) -> None:
        """Get the timestamp from a old style timestamp file."""
        if not old_timestamp_path.exists():
            return

        path = old_timestamp_path.parent
        mtime = old_timestamp_path.stat().st_mtime
        self._timestamps.set(path, mtime)

    def _import_old_parent_timestamps(self, path: Path) -> None:
        """Walk up to the root eating old style timestamps."""
        old_timestamp_path = path / _OLD_TIMESTAMPS_NAME
        self._add_old_timestamp(old_timestamp_path)
        if path.parent != path:
            self._import_old_parent_timestamps(path.parent)

    def _import_old_child_timestamps(self, path: Path) -> None:
        for root, dirnames, filenames in os.walk(path):
            root_path = Path(root)
            if _OLD_TIMESTAMPS_NAME in filenames:
                old_timestamp_path = root_path / _OLD_TIMESTAMPS_NAME
                self._add_old_timestamp(old_timestamp_path)
                self._timestamps._consumed_paths.add(old_timestamp_path)
            for dirname in dirnames:
                self._import_old_child_timestamps(root_path / dirname)

    def import_old_timestamps(self) -> None:
        """Import all old timestamps."""
        self._import_old_parent_timestamps(self._timestamps.dir)
        self._import_old_child_timestamps(self._timestamps.dir)

    def __init__(self, timestamps: Treestamps):
        """Hold new timestamp object."""
        self._timestamps = timestamps



"""Optimize comic archives."""
import shutil

from abc import ABCMeta, abstractmethod
from pathlib import Path
from typing import Optional, Union

from confuse.templates import AttrDict
from termcolor import cprint

from picopt.handlers.handler import Format, Handler, Metadata
from picopt.stats import ReportStats


class ContainerHandler(Handler, metaclass=ABCMeta):
    """Comic format class."""

    CONTAINER_DIR_SUFFIX: str = ".dir"
    CONVERT: bool = True

    @classmethod
    @abstractmethod
    def identify_format(cls, path: Path) -> Optional[Format]:
        """Return the format if this handler can handle this path."""
        pass

    @abstractmethod
    def unpack_into(self) -> None:
        """Unpack a container into a tmp dir to work on it's contents."""
        pass

    @abstractmethod
    def pack_into(self, working_path: Path) -> None:
        """Create a container from a tmp dir's contents."""
        pass

    def __init__(
        self,
        config: AttrDict,
        original_path: Path,
        format: Format,
        metadata: Metadata,
    ):
        """Unpack a container with a subclass's unpacker."""
        super().__init__(config, original_path, format, metadata)
        self.comment: Optional[bytes] = None
        self.tmp_container_dir: Path = Path(
            str(self.get_working_path()) + self.CONTAINER_DIR_SUFFIX
        )

    def unpack(self) -> Union[Handler, ReportStats]:
        """Create directory and unpack container."""
        try:
            if self.config.verbose:
                print(f"Extracting {self.original_path}...", end="")

            # create a clean tmpdir
            if self.tmp_container_dir.exists():
                shutil.rmtree(self.tmp_container_dir)
            self.tmp_container_dir.mkdir(parents=True)

            # extract archive into the tmpdir
            self.unpack_into()

            if self.config.verbose:
                print("done")
            return self
        except Exception as exc:
            return self.error(exc)

    def cleanup_after_optimize(self, working_path: Path) -> tuple[int, int]:
        """Clean up the temp dir as well as the old container."""
        if self.config.verbose:
            print(".", end="")
        shutil.rmtree(self.tmp_container_dir)

        if self.config.verbose:
            print(".", end="")
        bytes_count = super().cleanup_after_optimize(working_path)

        if self.config.verbose:
            print("done.")
        return bytes_count

    def repack(self) -> ReportStats:
        """Create a new container and clean up the tmp dir."""
        try:
            # archive into new filename
            new_path = self.get_working_path()
            if self.config.verbose:
                print(f"Repacking {new_path}", end="")
            self.pack_into(new_path)

            bytes_count = self.cleanup_after_optimize(new_path)
            report_stats = ReportStats(self.final_path, bytes_count=bytes_count)
            if self.config.verbose:
                report_stats.report(self.config.test)
            return report_stats
        except Exception as exc:
            cprint(f"ERROR: repack container: {exc}", "red")
            return self.error(exc)



"""Test comic format."""
import shutil

from picopt import PROGRAM_NAME, cli
from picopt.old_timestamps import _OLD_TIMESTAMPS_NAME
from tests import IMAGES_DIR, get_test_dir


__all__ = ()
TMP_ROOT = get_test_dir()
FN = "test_jpg.jpg"
SRC_JPG = IMAGES_DIR / FN
TMP_FN = str(TMP_ROOT / FN)
FNS = {
    FN: (97373, 87913),
}

DEFAULT_CONFIG = {
    "bigger": False,
    "convert_to": [],
    "formats": ["GIF", "PNG", "JPEG", "WEBP"],
    "keep_metadata": True,
    "ignore": [],
    "recurse": True,
    "symlinks": True,
}


class TestContainersDir:
    @staticmethod
    def _assert_sizes(index, root=TMP_ROOT):
        for name, sizes in FNS.items():
            path = root / name
            assert path.stat().st_size == sizes[index]

    def setup_method(self) -> None:
        self.teardown_method()
        TMP_ROOT.mkdir(exist_ok=True)
        shutil.copy(SRC_JPG, TMP_ROOT)
        self._assert_sizes(0)

    def teardown_method(self) -> None:
        if TMP_ROOT.exists():
            shutil.rmtree(TMP_ROOT)

    def test_old_timestamp_same_dir(self) -> None:
        old_ts_path = TMP_ROOT / _OLD_TIMESTAMPS_NAME
        old_ts_path.touch()
        args = (PROGRAM_NAME, "-rtvv", str(TMP_ROOT))
        res = cli.main(args)
        assert res
        assert not old_ts_path.exists()
        self._assert_sizes(0)

    def test_old_timestamp_child(self) -> None:
        child_root = TMP_ROOT / "child"
        child_root.mkdir(exist_ok=True)
        shutil.copy(SRC_JPG, child_root)
        old_ts_path = child_root / _OLD_TIMESTAMPS_NAME
        old_ts_path.touch()
        args = (PROGRAM_NAME, "-rtvv", str(child_root))
        res = cli.main(args)
        assert res
        assert not old_ts_path.exists()
        self._assert_sizes(0)

    def test_old_timestamp_parent(self) -> None:
        child_root = TMP_ROOT / "child"
        child_root.mkdir(exist_ok=True)
        shutil.copy(SRC_JPG, child_root)
        old_ts_path = TMP_ROOT / _OLD_TIMESTAMPS_NAME
        old_ts_path.touch()
        args = (PROGRAM_NAME, "-rtvv", str(child_root))
        res = cli.main(args)
        assert res
        assert old_ts_path.exists()
        self._assert_sizes(0)



"""Queue tasks."""
from abc import ABC
from dataclasses import dataclass, field
from pathlib import Path

from picopt.handlers.container import ContainerHandler
from picopt.stats import ReportStats


@dataclass
class DirResult:
    """Results from a directory."""

    path: Path
    results: list


@dataclass
class ContainerResult(DirResult):
    """Results from a container."""

    handler: ContainerHandler


@dataclass
class CompleteTask(ABC):
    """Generic abstract completion task."""

    pass


@dataclass
class CompleteDirTask(CompleteTask):
    """Task to compact timestamps."""

    path: Path


@dataclass
class CompleteContainerTask(CompleteTask):
    """Task to fire off repack once all container optimizations are done."""

    handler: ContainerHandler


@dataclass
class Totals:
    """Totals for final report."""

    bytes_in: int = 0
    bytes_out: int = 0
    errors: list[ReportStats] = field(default_factory=list)



"""Return a handler for a path."""
from pathlib import Path
from typing import Optional, Type

from confuse.templates import AttrDict
from PIL import Image, UnidentifiedImageError
from termcolor import cprint

from picopt.config import WEBP_CONVERTABLE_FORMATS
from picopt.handlers.container import ContainerHandler
from picopt.handlers.handler import Format, Handler, Metadata
from picopt.handlers.image import TIFF_FORMAT
from picopt.handlers.webp import WebPLossless
from picopt.handlers.zip import CBZ, EPub, Zip
from picopt.pillow.webp_lossless import is_lossless


_ALWAYS_LOSSLESS_FORMATS = WEBP_CONVERTABLE_FORMATS - set([TIFF_FORMAT])
_CONTAINER_HANDLERS: tuple[Type[ContainerHandler], ...] = (CBZ, Zip, EPub)


def _is_lossless(image_format: str, path: Path, info: dict) -> bool:
    """Determine if image format is lossless."""
    if image_format in _ALWAYS_LOSSLESS_FORMATS:
        lossless = True
    elif image_format == WebPLossless.OUTPUT_FORMAT:
        lossless = is_lossless(str(path))
    elif image_format == TIFF_FORMAT:
        lossless = info.get("compression") != "jpeg"
    else:
        lossless = False
    return lossless


def _get_image_format(
    path: Path, keep_metadata: bool
) -> tuple[Optional[Format], Metadata]:
    """Construct the image format with PIL."""
    format = None
    metadata = Metadata()
    n_frames = 1
    try:
        with Image.open(path, mode="r") as image:
            image.verify()
        image.close()  # for animated images
        with Image.open(path, mode="r") as image:
            image_format = image.format
            if image_format is None:
                raise UnidentifiedImageError("No image format")
            animated = getattr(image, "is_animated", False)
            info = image.info
            if keep_metadata and animated:
                n_frames = getattr(image, "n_frames", 1)
        image.close()  # for animated images
        if keep_metadata:
            exif = info.get("exif", b"")
            icc = info.get("icc_profile", "")
            metadata = Metadata(exif, icc, n_frames)
        lossless = _is_lossless(image_format, path, info)
        format = Format(image_format, lossless, animated)
    except UnidentifiedImageError:
        pass
    return format, metadata


def _get_container_format(path: Path) -> Optional[Format]:
    """Get the container format by querying each handler."""
    format = None
    for container_handler in _CONTAINER_HANDLERS:
        format = container_handler.identify_format(path)
        if format is not None:
            break
    return format


def _get_handler_class(config: AttrDict, key: str, format: Format) -> Type[Handler]:
    handler_classes = config._format_handlers.get(format)
    if handler_classes:
        handler_cls = handler_classes.get(key)
    else:
        handler_cls = None
    return handler_cls


def create_handler(
    config: AttrDict, path: Path, convert: bool = True
) -> Optional[Handler]:
    """Get the image format."""
    # This is the consumer of config._format_handlers
    format: Optional[Format] = None
    metadata: Metadata = Metadata()
    handler_cls: Optional[Type[Handler]] = None
    try:
        format, metadata = _get_image_format(path, config.keep_metadata)
        if not format:
            format = _get_container_format(path)

        if format:
            if convert:
                handler_cls = _get_handler_class(config, "convert", format)
            if not handler_cls:
                handler_cls = _get_handler_class(config, "native", format)
    except OSError as exc:
        cprint("WARNING: getting handler", str(exc), "yellow")

    if handler_cls and format is not None:
        handler = handler_cls(config, path, format, metadata)
    else:
        if config.verbose > 2 and not config.list_only:
            cprint(
                f"{path} is not an enabled image or container.", "white", attrs=["dark"]
            )
        handler = None
    return handler



"""PNG format."""
import shutil

from copy import copy
from pathlib import Path
from typing import Optional

from PIL.PngImagePlugin import PngImageFile
from termcolor import cprint

from picopt.handlers.handler import Format
from picopt.handlers.image import ImageHandler
from picopt.pillow.png_bit_depth import png_bit_depth


class Png(ImageHandler):
    """PNG format class."""

    BEST_ONLY: bool = False
    OUTPUT_FORMAT = PngImageFile.format
    OUTPUT_FORMAT_OBJ = Format(OUTPUT_FORMAT, True, False)
    PIL2_ARGS: dict[str, bool] = {"optimize": True}
    PROGRAMS: dict[str, Optional[str]] = ImageHandler.init_programs(
        ("pil2png", "optipng", "pngout")
    )
    PREFERRED_PROGRAM: str = "optipng"
    _OPTIPNG_ARGS = [PROGRAMS["optipng"], "-o5", "-fix", "-force", "-quiet"]
    _PNGOUT_ARGS = [PROGRAMS["pngout"], "-q", "-force", "-y"]

    def pil2png(self, old_path: Path, new_path: Path) -> Path:
        """Pillow png optimization."""
        return self.pil2native(old_path, new_path)

    def optipng(self, old_path: Path, new_path: Path) -> Path:
        """Run the external program optipng on the file."""
        shutil.copy2(old_path, new_path)
        args = copy(self._OPTIPNG_ARGS)
        if not self.config.keep_metadata:
            args += ["-strip", "all"]
        args += [str(new_path)]
        self.run_ext(tuple(args))
        return new_path

    def pngout(self, old_path: Path, new_path: Path) -> Path:
        """Run the external program pngout on the file."""
        depth = png_bit_depth(old_path)
        if depth in (16, None):
            cprint(
                f"Skipped pngout for {depth} bit PNG: {old_path}",
                "white",
                attrs=["dark"],
            )
            result = old_path
        else:
            args = tuple(self._PNGOUT_ARGS + [str(old_path), str(new_path)])
            self.run_ext(args)
            result = new_path
        return result
